import datetime
import dotenv
import os
import pkg_resources
import requests_cache
import yaml


__all__ = ['get_config']

BASE_DIR = pkg_resources.resource_filename('biosimulations_bigg', '.')


def get_config(
        source_api_endpoint='http://bigg.ucsd.edu/api/v2',
        source_model_file_endpoint='http://bigg.ucsd.edu/static',
        source_map_file_endpoint='http://bigg.ucsd.edu/escher_map_json',
        source_dirname=None,
        source_license_filename=None,
        sessions_dirname=None,
        final_dirname=None,
        curators_filename=None,
        issues_filename=None,
        status_filename=None,
        thumbnails_filename=None,
        extra_visualizations_filename=None,
        project_ids=None,
        first_project=0,
        max_projects=None,
        max_num_reactions=None,
        max_thumbnails=None,
        update_project_sources=False,
        update_combine_archives=False,
        update_simulations=False,
        update_simulation_runs=False,
        simulate_projects=True,
        publish_projects=True,
        entrez_delay=5.,
        bucket_endpoint=None,
        bucket_name=None,
        bucket_access_key_id=None,
        bucket_secret_access_key=None,
        biosimulations_api_client_id=None,
        biosimulations_api_client_secret=None,
        dry_run=False,
):
    """ Get a configuration

    Args:
        source_api_endpoint (obj:`str`, optional): endpoint for retrieving metadata about BiGG models
        source_model_file_endpoint (obj:`str`, optional): endpoint for retrieving files for BiGG models
        source_map_file_endpoint (obj:`str`, optional): endpoint for retrieving files for Escher visualizations
        source_dirname (obj:`str`, optional): directory where source models, metabolic flux maps, and thumbnails should be stored
        source_license_filename (obj:`str`, optional): path to BiGG license to copy into COMBINE/OMEX archives
        sessions_dirname (obj:`str`, optional): directory where cached HTTP sessions should be stored
        final_dirname (obj:`str`, optional): directory where created SED-ML, metadata, and COMBINE/OMEX archives should be stored
        curators_filename (obj:`str`, optional): path which describes the people who helped curator the repository
        issues_filename (obj:`str`, optional): path to issues which prevent some models from being imported
        status_filename (obj:`str`, optional): path to save the import status of each project
        thumbnails_filename (obj:`str`, optional): path to curated list of good thumbnails
        extra_visualizations_filename (obj:`str`, optional): path to curated list of additional Escher diagrams to use with projects
        project_ids (:obj:`list` of :obj:`str`, optional): ids of projects to import
        first_project (:obj:`int`, optional): iteration through projects at which to begin importing
        max_projects (:obj:`int`, optional): maximum number of projects to download, convert, execute, and submit; used for testing
        max_num_reactions (:obj:`int`, optional): maximum size model to import; used for testing
        max_thumbnails (:obj:`int`, optional): maximum number of thumbnails to use; used for testing
        update_project_sources (:obj:`bool`, optional): whether to update the source files for the projects; used for testing
        update_combine_archives (:obj:`bool`, optional): whether to update COMBINE archives even if they already exist; used for testing
        update_simulations (:obj:`bool`, optional): whether to re-run COMBINE archives even if they have already been run; used for testing
        update_simulation_runs (:obj:`bool`, optional): whether to update projects even if they have already been imported; used for testing
        simulate_projects (:obj:`bool`, optional): whether to simulate projects; used for testing
        publish_projects (:obj:`bool`, optional): whether to pushlish projects; used for testing
        entrez_delay (:obj:`float`, optional): delay in between Entrez queries
        bucket_endpoint (:obj:`str`, optional): endpoint for storage bucket
        bucket_name (:obj:`str`, optional): name of storage bucket
        bucket_access_key_id (:obj:`str`, optional): key id for storage bucket
        bucket_secret_access_key (:obj:`str`, optional): access key for storage bucket
        biosimulations_api_client_id (:obj:`str`, optional): id for client to the BioSimulations API
        biosimulations_api_client_secret (:obj:`str`, optional): secret for client to the BioSimulations API
        dry_run (:obj:`bool`, optional): whether to submit projects to BioSimulations or not; used for testing

    Returns:
        obj:`dict`: configuration
    """

    env = {
        **dotenv.dotenv_values("secret.env"),
        **os.environ,
    }

    if source_dirname is None:
        source_dirname = env.get('SOURCE_DIRNAME', os.path.join(BASE_DIR, 'source'))
    if source_license_filename is None:
        source_license_filename = env.get('SOURCE_LICENSE_FILENAME', os.path.join(BASE_DIR, 'source', 'LICENSE'))
    if sessions_dirname is None:
        sessions_dirname = env.get('SESSIONS_DIRNAME', os.path.join(BASE_DIR, 'source'))
    if final_dirname is None:
        final_dirname = env.get('FINAL_DIRNAME', os.path.join(BASE_DIR, 'final'))
    if curators_filename is None:
        curators_filename = env.get('CURATORS_FILENAME', os.path.join(BASE_DIR, 'final', 'curators.yml'))
    if issues_filename is None:
        issues_filename = env.get('ISSUES_FILENAME', os.path.join(BASE_DIR, 'final', 'issues.yml'))
    if status_filename is None:
        status_filename = env.get('STATUS_FILENAME', os.path.join(BASE_DIR, 'final', 'status.yml'))
    if thumbnails_filename is None:
        thumbnails_filename = env.get('THUMBNAILS_FILENAME', os.path.join(BASE_DIR, 'final', 'thumbnails.yml'))
    if extra_visualizations_filename is None:
        extra_visualizations_filename = env.get('EXTRA_VISUALIZATIONS_FILENAME',
                                                os.path.join(BASE_DIR, 'final', 'extra-visualizations.yml'))
    if bucket_endpoint is None:
        bucket_endpoint = env.get('BUCKET_ENDPOINT')
    if bucket_name is None:
        bucket_name = env.get('BUCKET_NAME')
    if bucket_access_key_id is None:
        bucket_access_key_id = env.get('BUCKET_ACCESS_KEY_ID')
    if bucket_secret_access_key is None:
        bucket_secret_access_key = env.get('BUCKET_SECRET_ACCESS_KEY')

    if biosimulations_api_client_id is None:
        biosimulations_api_client_id = env.get('BIOSIMULATIONS_API_CLIENT_ID')
    if biosimulations_api_client_secret is None:
        biosimulations_api_client_secret = env.get('BIOSIMULATIONS_API_CLIENT_SECRET')

    with open(curators_filename, 'r') as file:
        curators = yaml.load(file, Loader=yaml.Loader)

    return {
        'source_api_endpoint': source_api_endpoint,
        'source_model_file_endpoint': source_model_file_endpoint,
        'source_map_file_endpoint': source_map_file_endpoint,

        'source_models_dirname': os.path.join(source_dirname, 'models'),
        'source_visualizations_dirname': os.path.join(source_dirname, 'visualizations'),
        'source_thumbnails_dirname': os.path.join(source_dirname, 'thumbnails'),
        'source_license_filename': source_license_filename,

        'final_visualizations_dirname': os.path.join(final_dirname, 'visualizations'),
        'final_metadata_dirname': os.path.join(final_dirname, 'metadata'),
        'final_projects_dirname': os.path.join(final_dirname, 'projects'),
        'final_simulation_results_dirname': os.path.join(final_dirname, 'simulation-results'),

        'curators': curators,
        'issues_filename': issues_filename,
        'status_filename': status_filename,
        'thumbnails_filename': thumbnails_filename,
        'extra_visualizations_filename': extra_visualizations_filename,

        'source_session': requests_cache.CachedSession(
            os.path.join(sessions_dirname, 'source'),
            expire_after=datetime.timedelta(4 * 7)),
        'cross_ref_session': requests_cache.CachedSession(
            os.path.join(sessions_dirname, 'crossref'),
            expire_after=datetime.timedelta(4 * 7)),
        'pubmed_central_open_access_session': requests_cache.CachedSession(
            os.path.join(sessions_dirname, 'pubmed-central-open-access'),
            expire_after=datetime.timedelta(4 * 7)),

        'project_ids': project_ids,
        'first_project': first_project,
        'max_projects': max_projects,
        'max_num_reactions': max_num_reactions,
        'max_thumbnails': max_thumbnails,
        'update_project_sources': update_project_sources,
        'update_combine_archives': update_combine_archives,
        'update_simulations': update_simulations,
        'update_simulation_runs': update_simulation_runs,
        'simulate_projects': simulate_projects,
        'publish_projects': publish_projects,
        'entrez_delay': entrez_delay,
        'bucket_endpoint': bucket_endpoint,
        'bucket_name': bucket_name,
        'bucket_access_key_id': bucket_access_key_id,
        'bucket_secret_access_key': bucket_secret_access_key,
        'biosimulations_api_client_id': biosimulations_api_client_id,
        'biosimulations_api_client_secret': biosimulations_api_client_secret,
        'dry_run': dry_run,
    }



from .config import get_config
from .core import import_projects
from ._version import __version__
from biosimulators_utils.config import get_config as get_biosimulators_config
import biosimulators_utils.biosimulations.utils
import cement
import requests
import sys
import yaml


class BaseController(cement.Controller):
    """ Base controller for command line application """

    class Meta:
        label = 'base'
        description = "Utilities for publishing BiGG to BioSimulations"
        help = "Utilities for publishing BiGG to BioSimulations"
        arguments = [
            (['-v', '--version'], dict(
                action='version',
                version=__version__,
            )),
        ]

    @cement.ex(hide=True)
    def _default(self):
        self._parser.print_help()


class RunAndPublishProjectsController(cement.Controller):
    """ Publish projects from BiGG to BioSimulations

    * Download models
    * Download visualizations
    * Download metadata
    * Generate SED-ML files for models
    * Convert visualizations to Vega format
    * Expand taxonomy metadata using NCBI Taxonomy
    * Expand citation metadata using PubMed and CrossRef
    * Obtain thumbnail images using PubMed Central
    * Encode metadata into OMEX metadata files
    * Package project into COMBINE/OMEX archives
    * Submit simulation runs for archives to runBioSimulations
    * Publish simulation runs to BioSimulations
    """

    class Meta:
        label = 'run-projects-and-publish'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Publish projects from BiGG to BioSimulations"
        description = "Publish projects from BiGG to BioSimulations"
        arguments = [
            (
                ['--project'],
                dict(
                    type=str,
                    nargs='+',
                    action='extend',
                    help='Id of a project to run and publish. Used for testing.',
                ),
            ),
            (
                ['--first-project'],
                dict(
                    type=int,
                    default=1,
                    help='Iteration (1-based) through projects at which to begin importing. Used for testing.',
                ),
            ),
            (
                ['--max-projects'],
                dict(
                    type=int,
                    default=None,
                    help='Maximum number of projects to import. Used for testing.',
                ),
            ),
            (
                ['--max-num-reactions'],
                dict(
                    type=int,
                    default=None,
                    help='Maximum size model to import. Used for testing.',
                ),
            ),
            (
                ['--update-project-sources'],
                dict(
                    action='store_true',
                    help='If set, update the source files for the projects. Used for testing.'
                ),
            ),
            (
                ['--update-combine-archives'],
                dict(
                    action='store_true',
                    help='If set, update COMBINE/OMEX archive even if they have already been generated. Used for testing.'
                ),
            ),
            (
                ['--update-simulations'],
                dict(
                    action='store_true',
                    help='If set, re-run COMBINE/OMEX archives even if they have already been run. Used for testing.'
                ),
            ),
            (
                ['--update-simulation-runs'],
                dict(
                    action='store_true',
                    help='If set, update simulation runs even if they have already been submitted. Used for testing.'
                ),
            ),
            (
                ['--skip-simulation'],
                dict(
                    action='store_true',
                    help='If set, do not simulate projects. Used for testing.',
                ),
            ),
            (
                ['--skip-publication'],
                dict(
                    action='store_true',
                    help='If set, do not publish projects. Used for testing.',
                ),
            ),
            (
                ['--dry-run'],
                dict(
                    action='store_true',
                    help='If set, do not submit projects to BioSimulations. Used for testing.'
                ),
            ),
        ]

    @ cement.ex(hide=True)
    def _default(self):
        args = self.app.pargs

        config = get_config(project_ids=args.project,
                            first_project=args.first_project - 1,
                            max_projects=args.max_projects, max_num_reactions=args.max_num_reactions,
                            update_project_sources=args.update_project_sources,
                            update_combine_archives=args.update_combine_archives,
                            update_simulations=args.update_simulations,
                            update_simulation_runs=args.update_simulation_runs,
                            simulate_projects=not args.skip_simulation,
                            publish_projects=not args.skip_publication,
                            dry_run=args.dry_run)

        import_projects(config)


class PublishRunsController(cement.Controller):
    """ Publish runs of simulations of BiGG projects to BioSimulations """

    class Meta:
        label = 'publish-runs'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Publish runs of simulations to BioSimulations"
        description = "Publish runs of simulations of BiGG projects to BioSimulations"
        arguments = []

    @ cement.ex(hide=True)
    def _default(self):
        config = get_config()
        biosimulators_config = get_biosimulators_config()

        # read simulation runs
        projects_filename = config['status_filename']
        with open(projects_filename, 'r') as file:
            projects = yaml.load(file, Loader=yaml.Loader)

        # check status
        failures = []
        for id, project in projects.items():
            if project['runbiosimulationsId']:
                response = requests.get(biosimulators_config.BIOSIMULATIONS_API_ENDPOINT + 'runs/' + project['runbiosimulationsId'])
                response.raise_for_status()
                project['runbiosimulationsStatus'] = response.json()['status']
                if project['runbiosimulationsStatus'] != 'SUCCEEDED':
                    failures.append('{}: {}'.format(id, project['runbiosimulationsStatus']))
            else:
                failures.append('{}: {}'.format(id, 'not submitted'))
        if failures:
            raise SystemExit('{} simulation runs did not succeed:\n  {}'.format(len(failures), '\n  '.join(sorted(failures))))

        # login to publish projects
        auth_headers = {
            'Authorization': biosimulators_utils.biosimulations.utils.get_authorization_for_client(
                config['biosimulations_api_client_id'], config['biosimulations_api_client_secret'])
        }

        # publish projects
        print('Publishing or updating {} projects ...'.format(len(projects)))
        for i_project, (id, project) in enumerate(projects.items()):
            print('  {}: {} ... '.format(i_project + 1, id), end='')
            sys.stdout.flush()

            endpoint = biosimulators_config.BIOSIMULATIONS_API_ENDPOINT + 'projects/' + id

            response = requests.get(endpoint)

            if response.status_code == 200:
                if response.json()['simulationRun'] == project['runbiosimulationsId']:
                    api_method = None
                    print('already up to date. ', end='')
                    sys.stdout.flush()

                else:
                    api_method = requests.put
                    print('updating ... ', end='')
                    sys.stdout.flush()

            else:
                api_method = requests.post
                print('publishing ... ', end='')
                sys.stdout.flush()

            if api_method:
                response = api_method(endpoint,
                                      headers=auth_headers,
                                      json={
                                          'id': id,
                                          'simulationRun': project['runbiosimulationsId']
                                      })
                response.raise_for_status()

            print('done.')
        print('')

        # print message
        print('All {} projects were successfully published or updated'.format(len(projects)))


class VerifyPublicationController(cement.Controller):
    """ Verify that projects from BiGG have been successfully published to BioSimulations """

    class Meta:
        label = 'verify-publication'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = "Verify that projects have been published to BioSimulations"
        description = "Verify that projects from BiGG have been successfully published to BioSimulations"
        arguments = []

    @ cement.ex(hide=True)
    def _default(self):
        config = get_config()
        biosimulators_config = get_biosimulators_config()

        # read BiGG projects
        bigg_projects_filename = config['status_filename']
        with open(bigg_projects_filename, 'r') as file:
            bigg_projects = yaml.load(file, Loader=yaml.Loader)

        # get BioSimulations projects
        biosimulations_api_endpoint = biosimulators_config.BIOSIMULATIONS_API_ENDPOINT
        response = requests.get(biosimulations_api_endpoint + 'projects')
        response.raise_for_status()
        biosimulations_projects = {
            project['id']: project
            for project in response.json()
        }

        # check all BiGG projects were published
        errors = []
        for bigg_project_id, bigg_project in bigg_projects.items():
            if bigg_project_id not in biosimulations_projects:
                if bigg_project['runbiosimulationsId']:
                    biosimulations_api_endpoint = biosimulators_config.BIOSIMULATIONS_API_ENDPOINT
                    response = requests.get(biosimulations_api_endpoint + 'runs/{}'.format(bigg_project['runbiosimulationsId']))
                    response.raise_for_status()
                    run_status = response.json()['status']
                else:
                    run_status = 'not submitted'
                errors.append('{}: has not been published. The status of run `{}` is `{}`.'.format(
                    bigg_project_id, bigg_project['runbiosimulationsId'], run_status))
            elif biosimulations_projects[bigg_project_id]['simulationRun'] != bigg_project['runbiosimulationsId']:
                biosimulations_api_endpoint = biosimulators_config.BIOSIMULATIONS_API_ENDPOINT
                url = biosimulations_api_endpoint + 'projects/{}'.format(bigg_project_id)
                response = requests.get(url)
                try:
                    response.raise_for_status()
                    owner = response.json().get('owner', {}).get('name', None)

                    if owner != 'BiGG':
                        reason = 'Project id has already been claimed.'
                    else:
                        reason = 'Project is published with run `{}`.'.format(response.json()['simulationRun'])
                except requests.exceptions.RequestException:
                    reason = 'Project could not be found'

                errors.append('{}: not published as run {}. {}'.format(bigg_project_id, bigg_project['runbiosimulationsId'], reason))

        # print message
        if errors:
            msg = '{} projects have been successfully published.\n\n'.format(len(bigg_projects) - len(errors))
            msg += '{} projects have not been successfully published:\n  {}'.format(len(errors), '\n  '.join(errors))
            raise SystemExit(msg)

        else:
            print('All {} projects have been successfully published.'.format(len(bigg_projects)))


class App(cement.App):
    """ Command line application """
    class Meta:
        label = 'biosimulations-bigg'
        base_controller = 'base'
        handlers = [
            BaseController,
            RunAndPublishProjectsController,
            PublishRunsController,
            VerifyPublicationController,
        ]


def main():
    with App() as app:
        app.run()



import re
import setuptools
import subprocess
import sys
try:
    result = subprocess.run(
        [sys.executable, "-m", "pip", "show", "pkg_utils"],
        check=True, capture_output=True)
    match = re.search(r'\nVersion: (.*?)\n', result.stdout.decode(), re.DOTALL)
    assert match and tuple(match.group(1).split('.')) >= ('0', '0', '5')
except (subprocess.CalledProcessError, AssertionError):
    subprocess.run(
        [sys.executable, "-m", "pip", "install", "-U", "pkg_utils"],
        check=True)
import os
import pkg_utils

name = 'biosimulations_bigg'
dirname = os.path.dirname(__file__)
package_data = {
    name: [
        os.path.join('source', 'LICENSE'),
        os.path.join('final', '*.yml'),
    ],
}

# get package metadata
md = pkg_utils.get_package_metadata(dirname, name, package_data_filename_patterns=package_data)

# install package
setuptools.setup(
    name='biosimulations-bigg',
    version=md.version,
    description=(
        "Command-line program for publishing the BiGG model repository to BioSimulations."
    ),
    long_description=md.long_description,
    url="https://github.com/biosimulations/biosimulations-bigg",
    download_url='https://github.com/biosimulations/biosimulations-bigg',
    author='Center for Reproducible Biomedical Modeling',
    author_email="info@biosimulations.org",
    license="MIT",
    keywords=[
        'metabolism',
        'biochemical network',
        'cell biology',
        'systems biology',
        'computational biology',
        'mathematical modeling',
        'numerical simulation',
        'flux balance analysis',
        'FBA',
        'Systems Biology Markup Language',
        'SBML',
        'BiGG',
    ],
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    package_data=md.package_data,
    install_requires=md.install_requires,
    extras_require=md.extras_require,
    tests_require=md.tests_require,
    dependency_links=md.dependency_links,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    entry_points={
        'console_scripts': [
            'biosimulations-bigg = biosimulations_bigg.__main__:main',
        ],
    },
)



from Bio import Entrez
from biosimulators_utils.combine.data_model import CombineArchive, CombineArchiveContent, CombineArchiveContentFormat
from biosimulators_utils.combine.io import CombineArchiveWriter
from biosimulators_utils.config import Config
from biosimulators_utils.omex_meta.data_model import BIOSIMULATIONS_ROOT_URI_FORMAT, OmexMetadataOutputFormat
from biosimulators_utils.omex_meta.io import BiosimulationsOmexMetaWriter, BiosimulationsOmexMetaReader
# from biosimulators_utils.omex_meta.utils import build_omex_meta_file_for_model
from biosimulators_utils.ref.data_model import Reference, JournalArticle, PubMedCentralOpenAccesGraphic  # noqa: F401
from biosimulators_utils.ref.utils import get_reference, get_pubmed_central_open_access_graphics
from biosimulators_utils.sedml.data_model import (
    SedDocument, Model, ModelLanguage, SteadyStateSimulation,
    Task, DataGenerator, Report, DataSet)
from biosimulators_utils.sedml.io import SedmlSimulationWriter
from biosimulators_utils.sedml.model_utils import get_parameters_variables_outputs_for_simulation
from biosimulators_utils.viz.vega.escher import escher_to_vega
from biosimulators_utils.warnings import BioSimulatorsWarning
from unittest import mock
import biosimulators_cobrapy
import biosimulators_utils.biosimulations.utils
import boto3
import copy
import dataclasses
import datetime
import dateutil.parser
import dotenv
import os
import re
import shutil
import time
import warnings
import yaml

env = {
    **dotenv.dotenv_values("config.env"),
    **os.environ,
}

Entrez.email = env.get('ENTREZ_EMAIL', None)

__all__ = ['import_projects']


def get_models(config):
    """ Get a list of the models in the source database

    Args:
        config (:obj:`dict`): configuration

    Returns:
        :obj:`list` of :obj:`dict`: models
    """
    response = config['source_session'].get(config['source_api_endpoint'] + '/models')
    response.raise_for_status()
    models = response.json()['results']
    models.sort(key=lambda model: model['bigg_id'])
    return models


def get_model_details(model, last_updated, config):
    """ Get the details of a model from the source database and download the associated files

    Args:
        model (:obj:`dict`): model
        last_updated (:obj:`datetime.datetime`): timestamp when the model was last updated
        config (:obj:`dict`): configuration

    Returns:
        :obj:`dict`: detailed information about the model
    """
    # get information about the model
    response = config['source_session'].get(config['source_api_endpoint'] + '/models/' + model['bigg_id'])
    response.raise_for_status()
    model_detail = response.json()

    download_files = (
        last_updated is None
        or config['update_project_sources']
        or (dateutil.parser.parse(model_detail['last_updated']) + datetime.timedelta(1)) > last_updated
    )

    if download_files:
        # download the file for the model
        model_filename = os.path.join(config['source_models_dirname'], model['bigg_id'] + '.xml')
        response = config['source_session'].get(config['source_model_file_endpoint'] + '/models/{}.xml'.format(model['bigg_id']))
        response.raise_for_status()
        with open(model_filename, 'wb') as file:
            file.write(response.content)

        # download flux map visualizations associated with the model
        for escher_map in model_detail['escher_maps']:
            map_name = escher_map['map_name']
            standardized_map_name = re.sub(r'[^a-zA-Z0-9\.]', '-', map_name)
            escher_filename = os.path.join(config['source_visualizations_dirname'], standardized_map_name + '.json')
            response = config['source_session'].get(config['source_map_file_endpoint'] + '/' + map_name)
            response.raise_for_status()
            with open(escher_filename, 'wb') as file:
                file.write(response.content)

    # return the details of the model
    return model_detail


def get_metadata_for_model(model_detail, config):
    """ Get additional metadata about a model

    * NCBI Taxonomy id of the organism
    * PubMed id, PubMed Central id and DOI for the reference
    * Open access figures for the reference

    Args:
        model_detail (:obj:`dict`): information about a model
        config (:obj:`dict`): configuration

    Returns:
        :obj:`tuple`:

            * :obj:`dict`: NCBI taxonomy identifier and name
            * :obj:`dict`: Genome identifier and name
            * :obj:`Reference`: structured information about the reference
            * :obj:`list` of :obj:`PubMedCentralOpenAccesGraphic`: figures of the reference
    """
    metadata_filename = os.path.join(config['final_metadata_dirname'], model_detail['model_bigg_id'] + '.yml')
    if os.path.isfile(metadata_filename):
        with open(metadata_filename, 'r') as file:
            metadata = yaml.load(file, Loader=yaml.Loader)
        taxon = metadata.get('taxon', None)
        encodes = metadata.get('encodes', None)
        reference = metadata.get('reference', None)
        thumbnails = metadata.get('thumbnails', None)

        if reference:
            reference = JournalArticle(**reference)
        if thumbnails:
            for thumbnail in thumbnails:
                thumbnail['filename'] = os.path.join(config['source_thumbnails_dirname'], thumbnail['filename'])

            thumbnails = [PubMedCentralOpenAccesGraphic(**thumbnail) for thumbnail in thumbnails]
    else:
        taxon = None
        encodes = None
        reference = None
        thumbnails = None

    if (
        taxon is not None
        and encodes is not None
        and reference is not None
        and thumbnails is not None
    ):
        return taxon, encodes, reference, thumbnails

    # NCBI id for organism
    if taxon is None or encodes is None:
        time.sleep(config['entrez_delay'])
        handle = Entrez.esearch(db="nucleotide", term='{}[Assembly] OR {}[Primary Accession]'.format(
            model_detail['genome_name'], model_detail['genome_name']), retmax=1, retmode="xml")
        record = Entrez.read(handle)
        handle.close()
        if len(record["IdList"]) > 0:
            nucleotide_id = record["IdList"][0]

            time.sleep(config['entrez_delay'])
            handle = Entrez.esummary(db="nucleotide", id=nucleotide_id, retmode="xml")
            records = list(Entrez.parse(handle))
            handle.close()
            assert len(records) == 1

            encodes = {
                'uri': 'https://www.ncbi.nlm.nih.gov/nuccore/' + str(records[0]['Id']),
                'label': str(records[0]['Title']),
            }

            taxon_id = int(records[0]['TaxId'].real)

        else:
            time.sleep(config['entrez_delay'])
            handle = Entrez.esearch(db="assembly", term='{}'.format(
                model_detail['genome_name']), retmax=1, retmode="xml")
            record = Entrez.read(handle)
            handle.close()
            if len(record["IdList"]) == 0:
                raise ValueError('Genome assembly `{}` could not be found for model `{}`'.format(
                    model_detail['genome_name'], model_detail['model_bigg_id']))

            assembly_id = str(record["IdList"][0])

            time.sleep(config['entrez_delay'])
            handle = Entrez.esummary(db="assembly", id=assembly_id, retmode="xml")
            record = Entrez.read(handle)['DocumentSummarySet']['DocumentSummary'][0]
            handle.close()

            encodes = {
                'uri': 'https://www.ncbi.nlm.nih.gov/assembly/' + assembly_id,
                'label': '{} genome assembly {}'.format(record['Organism'], record['AssemblyName']),
            }

            taxon_id = int(record['SpeciesTaxid'])

        time.sleep(config['entrez_delay'])
        handle = Entrez.esummary(db="taxonomy", id=taxon_id, retmode="xml")
        record = Entrez.read(handle)
        assert len(record) == 1
        handle.close()

        taxon = {
            'id': taxon_id,
            'name': str(record[0]['ScientificName']),
        }

    # Citation information for the associated publication
    if reference is None:
        reference = get_reference(
            model_detail['reference_id'] or None if model_detail['reference_type'] == 'pmid' else None,
            model_detail['reference_id'] or None if model_detail['reference_type'] == 'doi' else None,
            cross_ref_session=config['cross_ref_session'],
        )

    # Figures for the associated publication from open-access subset of PubMed Central
    if thumbnails is None:
        if reference and reference.pubmed_central_id:
            thumbnails = get_pubmed_central_open_access_graphics(
                reference.pubmed_central_id,
                os.path.join(config['source_thumbnails_dirname'], reference.pubmed_central_id),
                session=config['pubmed_central_open_access_session'],
            )
        else:
            thumbnails = []

    # save metadata
    metadata = {
        'taxon': taxon,
        'encodes': encodes,
        'reference': reference.__dict__,
        'thumbnails': [dataclasses.asdict(thumbnail) for thumbnail in thumbnails],
    }
    for thumbnail in metadata['thumbnails']:
        thumbnail['filename'] = os.path.relpath(thumbnail['filename'], config['source_thumbnails_dirname'])

    with open(metadata_filename, 'w') as file:
        file.write(yaml.dump(metadata))

    return (taxon, encodes, reference, thumbnails)


def export_project_metadata_for_model_to_omex_metadata(model_detail, taxon, encodes, reference, thumbnails, metadata_filename, config):
    """ Export metadata about a model to an OMEX metadata RDF-XML file

    Args:
        model_detail (:obj:`str`): information about the model
        taxon (:obj:`dict`): NCBI taxonomy identifier and name
        encodes (:obj:`dict`): Genome identifier and name
        reference (:obj:`Reference`): structured information about the reference
        thumbnails (:obj:`list` of :obj:`PubMedCentralOpenAccesGraphic`): figures of the reference
        metadata_filename (:obj:`str`): path to save metadata
        config (:obj:`dict`): configuration
    """
    created = reference.date
    last_updated = dateutil.parser.parse(model_detail['last_updated'])
    metadata = [{
        "uri": '.',
        "combine_archive_uri": BIOSIMULATIONS_ROOT_URI_FORMAT.format(model_detail['model_bigg_id']),
        'title': '{}: {} metabolism'.format(model_detail['model_bigg_id'], taxon['name']),
        'abstract': 'Flux balance analysis model of the metabolism of {}.'.format(taxon['name']),
        'keywords': [
            'metabolism',
        ],
        'description': None,
        'taxa': [
            {
                'uri': 'http://identifiers.org/taxonomy:{}'.format(taxon['id']),
                'label': taxon['name'],
            },
        ],
        'encodes': [
            {
                'uri': 'http://identifiers.org/GO:0008152',
                'label': 'metabolic process',
            },
            encodes,
        ],
        'thumbnails': [thumbnail.location for thumbnail in thumbnails],
        'sources': [],
        'predecessors': [],
        'successors': [],
        'see_also': [],
        'creators': [
            {
                'uri': None,
                'label': author,
            } for author in reference.authors
        ],
        'contributors': config['curators'],
        'identifiers': [
            {
                'uri': 'http://identifiers.org/bigg.model:{}'.format(model_detail['model_bigg_id']),
                'label': 'bigg.model:{}'.format(model_detail['model_bigg_id']),
            },
        ],
        'citations': [
            {
                'uri': (
                    'http://identifiers.org/doi:' + reference.doi
                    if reference.doi else
                    'http://identifiers.org/pubmed:' + reference.pubmed_id
                ),
                'label': reference.get_citation(),
            },
        ],
        'license': {
            'uri': 'http://bigg.ucsd.edu/license',
            'label': 'BiGG',
        },
        'funders': [],
        'created': created,
        'modified': [
            '{}-{:02d}-{:02d}'.format(last_updated.year, last_updated.month, last_updated.day),
        ],
        'other': [],
    }]
    config = Config(OMEX_METADATA_OUTPUT_FORMAT=OmexMetadataOutputFormat.rdfxml)
    BiosimulationsOmexMetaWriter().run(metadata, metadata_filename, config=config)
    _, errors, warnings = BiosimulationsOmexMetaReader().run(metadata_filename)
    assert not errors


def build_combine_archive_for_model(model_filename, archive_dirname, archive_filename, extra_contents):
    params, sims, vars, outputs = get_parameters_variables_outputs_for_simulation(
        model_filename, ModelLanguage.SBML, SteadyStateSimulation, native_ids=True)

    obj_vars = list(filter(lambda var: var.target.startswith('/sbml:sbml/sbml:model/fbc:listOfObjectives/'), vars))
    rxn_flux_vars = list(filter(lambda var: var.target.startswith('/sbml:sbml/sbml:model/sbml:listOfReactions/'), vars))

    sedml_doc = SedDocument()
    model = Model(
        id='model',
        source=os.path.basename(model_filename),
        language=ModelLanguage.SBML.value,
        changes=params,
    )
    sedml_doc.models.append(model)
    sim = sims[0]
    sedml_doc.simulations.append(sim)

    task = Task(
        id='task',
        model=model,
        simulation=sim,
    )
    sedml_doc.tasks.append(task)

    report = Report(
        id='objective',
        name='Objective',
    )
    sedml_doc.outputs.append(report)
    for var in obj_vars:
        var_id = var.id
        var_name = var.name

        var.id = 'variable_' + var_id
        var.name = None

        var.task = task
        data_gen = DataGenerator(
            id='data_generator_{}'.format(var_id),
            variables=[var],
            math=var.id,
        )
        sedml_doc.data_generators.append(data_gen)
        report.data_sets.append(DataSet(
            id=var_id,
            label=var_id,
            name=var_name,
            data_generator=data_gen,
        ))

    report = Report(
        id='reaction_fluxes',
        name='Reaction fluxes',
    )
    sedml_doc.outputs.append(report)
    for var in rxn_flux_vars:
        var_id = var.id
        var_name = var.name

        var.id = 'variable_' + var_id
        var.name = None

        var.task = task
        data_gen = DataGenerator(
            id='data_generator_{}'.format(var_id),
            variables=[var],
            math=var.id,
        )
        sedml_doc.data_generators.append(data_gen)
        report.data_sets.append(DataSet(
            id=var_id,
            label=var_id[2:] if var_id.startswith('R_') else var_id,
            name=var_name if len(rxn_flux_vars) < 4000 else None,
            data_generator=data_gen,
        ))

    # make directory for archive
    if os.path.isdir(archive_dirname):
        shutil.rmtree(archive_dirname)
    os.makedirs(archive_dirname)
    shutil.copyfile(model_filename, os.path.join(archive_dirname, os.path.basename(model_filename)))

    SedmlSimulationWriter().run(sedml_doc, os.path.join(archive_dirname, 'simulation.sedml'))

    # form a description of the archive
    archive = CombineArchive()
    archive.contents.append(CombineArchiveContent(
        location=os.path.basename(model_filename),
        format=CombineArchiveContentFormat.SBML.value,
    ))
    archive.contents.append(CombineArchiveContent(
        location='simulation.sedml',
        format=CombineArchiveContentFormat.SED_ML.value,
        master=True,
    ))
    for local_path, extra_content in extra_contents.items():
        shutil.copyfile(local_path, os.path.join(archive_dirname, extra_content.location))
        archive.contents.append(extra_content)

    # save archive to file
    CombineArchiveWriter().run(archive, archive_dirname, archive_filename)


def import_projects(config):
    """ Download the source database, convert into COMBINE/OMEX archives, simulate the archives, and submit them to BioSimulations

    Args:
        config (:obj:`dict`): configuration
    """

    # create directories for source files, thumbnails, projects, and simulation results
    if not os.path.isdir(config['source_models_dirname']):
        os.makedirs(config['source_models_dirname'])
    if not os.path.isdir(config['source_visualizations_dirname']):
        os.makedirs(config['source_visualizations_dirname'])
    if not os.path.isdir(config['source_thumbnails_dirname']):
        os.makedirs(config['source_thumbnails_dirname'])

    if not os.path.isdir(config['final_visualizations_dirname']):
        os.makedirs(config['final_visualizations_dirname'])
    if not os.path.isdir(config['final_metadata_dirname']):
        os.makedirs(config['final_metadata_dirname'])
    if not os.path.isdir(config['final_projects_dirname']):
        os.makedirs(config['final_projects_dirname'])
    if not os.path.isdir(config['final_simulation_results_dirname']):
        os.makedirs(config['final_simulation_results_dirname'])

    # read import status file
    if os.path.isfile(config['status_filename']):
        with open(config['status_filename'], 'r') as file:
            status = yaml.load(file, Loader=yaml.Loader)
    else:
        status = {}

    # read import issues file
    with open(config['issues_filename'], 'r') as file:
        issues = yaml.load(file, Loader=yaml.Loader)

    # read thumbnails file
    if os.path.isfile(config['thumbnails_filename']):
        with open(config['thumbnails_filename'], 'r') as file:
            thumbnails_curation = yaml.load(file, Loader=yaml.Loader)
        for model in thumbnails_curation.values():
            for thumbnail in model:
                thumbnail['filename'] = os.path.join(config['source_thumbnails_dirname'], thumbnail['filename'])
    else:
        thumbnails_curation = {}

    # read extra visualizations file
    if os.path.isfile(config['extra_visualizations_filename']):
        with open(config['extra_visualizations_filename'], 'r') as file:
            extra_visualizations_curation = yaml.load(file, Loader=yaml.Loader)
    else:
        extra_visualizations_curation = {}

    # get a list of all models available in the source database
    models = get_models(config)

    # filter to selected projects
    if config['project_ids'] is not None:
        models = list(filter(lambda model: model['bigg_id'] in config['project_ids'], models))

    # limit the models to import by number of reactions (used for testing)
    if config['max_num_reactions'] is not None:
        models = list(filter(lambda model: model['reaction_count'] < config['max_num_reactions'], models))

    # limit the number of models to import
    models = models[config['first_project']:]
    models = models[0:config['max_projects']]

    # get the details of each model
    model_details = []
    update_times = {}
    for i_model, model in enumerate(models):
        print('Retrieving {} of {}: {} ...'.format(i_model + 1, len(models), model['bigg_id']))

        # update status
        update_times[model['bigg_id']] = datetime.datetime.utcnow()

        # get the details of the model and download it from the source database
        last_updated = status.get(model['bigg_id'], {}).get('updated', None)
        if last_updated:
            last_updated = dateutil.parser.parse(last_updated)

        model_detail = get_model_details(model, last_updated, config)
        model_details.append(model_detail)
    models = model_details

    # filter out models that don't need to be imported because they've already been imported and haven't been updated
    if not config['update_simulation_runs']:
        models = list(filter(
            lambda model:
            (
                model['model_bigg_id'] not in status
                or not status[model['model_bigg_id']]['runbiosimulationsId']
                or (
                    (dateutil.parser.parse(model['last_updated']) + datetime.timedelta(1))
                    > dateutil.parser.parse(status[model['model_bigg_id']]['updated'])
                )
            ),
            models
        ))

    # filter out models with issues
    models = list(filter(lambda model: model['model_bigg_id'] not in issues, models))

    # get S3 bucket to save archives
    s3 = boto3.resource('s3',
                        endpoint_url=config['bucket_endpoint'],
                        aws_access_key_id=config['bucket_access_key_id'],
                        aws_secret_access_key=config['bucket_secret_access_key'])
    bucket = s3.Bucket(config['bucket_name'])

    # get authorization for BioSimulations API
    auth = biosimulators_utils.biosimulations.utils.get_authorization_for_client(
        config['biosimulations_api_client_id'], config['biosimulations_api_client_secret'])

    # download models, convert them to COMBINE/OMEX archives, simulate them, and deposit them to the BioSimulations database
    for i_model, model in enumerate(models):
        project_filename = os.path.join(config['final_projects_dirname'], model['model_bigg_id'] + '.omex')
        if not os.path.isfile(project_filename) or config['update_combine_archives']:
            model_filename = os.path.join(config['source_models_dirname'], model['model_bigg_id'] + '.xml')

            # get additional metadata about the model
            print('Getting metadata for {} of {}: {}'.format(i_model + 1, len(models), model['model_bigg_id']))
            taxon, encodes, reference, thumbnails = get_metadata_for_model(model, config)

            # filter out disabled thumbnails
            if thumbnails_curation.get(model['model_bigg_id'], []):
                thumbnails = []
                for thumbnail in thumbnails_curation[model['model_bigg_id']]:
                    if thumbnail['enabled']:
                        thumbnails.append(PubMedCentralOpenAccesGraphic(
                            id=thumbnail['id'],
                            label=thumbnail['label'],
                            filename=thumbnail['filename'],
                        ))
            else:
                thumbnails_curation[model['model_bigg_id']] = [
                    {
                        'id': thumbnail.id,
                        'label': thumbnail.label,
                        'filename': thumbnail.filename,
                        'enabled': True,
                    }
                    for thumbnail in thumbnails
                ]
            thumbnails = thumbnails[0:config['max_thumbnails']]
            for thumbnail in thumbnails:
                thumbnail.location = reference.pubmed_central_id + '-' + os.path.basename(thumbnail.id) + '.jpg'
                thumbnail.format = CombineArchiveContentFormat.JPEG

            # convert Escher map to Vega and add to thumbnails
            escher_maps = model['escher_maps'] + sorted((
                {'map_name': name} for name in set(extra_visualizations_curation.get(model['model_bigg_id'], [])).difference(
                    set(map['map_name'] for map in model['escher_maps'])
                )),
                key=lambda map: map['map_name'],
            )

            for escher_map in escher_maps:
                map_name = escher_map['map_name']
                standardized_map_name = re.sub(r'[^a-zA-Z0-9\.]', '-', map_name)
                escher_filename = os.path.join(config['source_visualizations_dirname'], standardized_map_name + '.json')

                vega_filename = os.path.join(config['final_visualizations_dirname'], standardized_map_name + '.vg.json')
                if not os.path.isfile(vega_filename):
                    reaction_fluxes_data_set = {
                        'sedmlUri': ['simulation.sedml', 'reaction_fluxes'],
                    }
                    escher_to_vega(reaction_fluxes_data_set, escher_filename, vega_filename)

                png_filename = os.path.join(config['source_visualizations_dirname'], map_name + '.png')
                if os.path.isfile(png_filename):
                    thumbnails.append(mock.Mock(
                        filename=png_filename,
                        location=standardized_map_name + '.png',
                        format=CombineArchiveContentFormat.PNG,
                    ))

            # export metadata to RDF
            print('Exporting project metadata for {} of {}: {}'.format(i_model + 1, len(models), model['model_bigg_id']))
            project_metadata_filename = os.path.join(config['final_metadata_dirname'], model['model_bigg_id'] + '.rdf')
            export_project_metadata_for_model_to_omex_metadata(model, taxon, encodes, reference, thumbnails,
                                                               project_metadata_filename, config)

            # print('Exporting model metadata for {} of {}: {}'.format(i_model + 1, len(models), model['model_bigg_id']))
            # model_metadata_filename = os.path.join(config['final_metadata_dirname'], model['model_bigg_id'] + '-omex-metadata.rdf')
            # build_omex_meta_file_for_model(model_filename, model_metadata_filename, metadata_format=OmexMetaOutputFormat.rdfxml_abbrev)

            # package model into COMBINE/OMEX archive
            print('Converting model {} of {}: {} ...'.format(i_model + 1, len(models), model['model_bigg_id']))

            extra_contents = {}
            extra_contents[project_metadata_filename] = CombineArchiveContent(
                location='metadata.rdf',
                format=CombineArchiveContentFormat.OMEX_METADATA,
            )
            # extra_contents[model_metadata_filename] = CombineArchiveContent(
            #     location=model['model_bigg_id'] + '.rdf',
            #     format=CombineArchiveContentFormat.OMEX_METADATA,
            # )
            extra_contents[config['source_license_filename']] = CombineArchiveContent(
                location='LICENSE',
                format=CombineArchiveContentFormat.TEXT,
            )
            for escher_map in escher_maps:
                map_name = escher_map['map_name']
                standardized_map_name = re.sub(r'[^a-zA-Z0-9\.]', '-', map_name)
                escher_filename = os.path.join(config['source_visualizations_dirname'], standardized_map_name + '.json')
                vega_filename = os.path.join(config['final_visualizations_dirname'], standardized_map_name + '.vg.json')
                extra_contents[escher_filename] = CombineArchiveContent(
                    location=standardized_map_name + '.escher.json',
                    format=CombineArchiveContentFormat.Escher,
                )
                extra_contents[vega_filename] = CombineArchiveContent(
                    location=standardized_map_name + '.vg.json',
                    format=CombineArchiveContentFormat.Vega,
                )
            for thumbnail in thumbnails:
                extra_contents[thumbnail.filename] = CombineArchiveContent(
                    location=thumbnail.location,
                    format=thumbnail.format,
                )

            project_dirname = os.path.join(config['final_projects_dirname'], model['model_bigg_id'])
            project_filename = os.path.join(config['final_projects_dirname'], model['model_bigg_id'] + '.omex')

            build_combine_archive_for_model(model_filename, project_dirname, project_filename, extra_contents=extra_contents)

        # simulate COMBINE/OMEX archives
        prev_objective = status.get(model['model_bigg_id'], {}).get('objective', None)
        prev_duration = status.get(model['model_bigg_id'], {}).get('duration', None)

        if config['simulate_projects'] and (
            config['update_combine_archives']
            or config['update_simulations']
            or prev_objective is None
        ):
            print('Simulating model {} of {}: {} ...'.format(i_model + 1, len(models), model['model_bigg_id']))
            out_dirname = os.path.join(config['final_simulation_results_dirname'], model['model_bigg_id'])
            biosimulators_utils_config = Config(COLLECT_COMBINE_ARCHIVE_RESULTS=True)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", BioSimulatorsWarning)
                results, log = biosimulators_cobrapy.exec_sedml_docs_in_combine_archive(
                    project_filename, out_dirname, config=biosimulators_utils_config)
            if log.exception:
                print('Simulation of `{}` failed'.format(model['model_bigg_id']))
                raise log.exception
            objective = results['simulation.sedml']['objective']['obj'].tolist()
            print('  {}: Objective: {}'.format(model['model_bigg_id'], objective))
            if objective <= 0:
                raise ValueError('`{}` is not a meaningful simulation.'.format(model['model_bigg_id']))
            duration = log.duration
        else:
            objective = prev_objective
            duration = prev_duration

        # submit COMBINE/OMEX archive to BioSimulations
        if config['dry_run']:
            runbiosimulations_id = status.get(model['model_bigg_id'], {}).get('runbiosimulationsId', None)
            updated = status.get(model['model_bigg_id'], {}).get('updated', None)
        else:
            print('Submitting model {} of {}: {} ...'.format(i_model + 1, len(models), model['model_bigg_id']))

            name = model['model_bigg_id']
            if config['publish_projects']:
                project_id = name
            else:
                project_id = None

            project_bucket_key = '{}.omex'.format(model['model_bigg_id'])
            bucket.upload_file(project_filename, project_bucket_key)
            project_url = '{}/{}/{}'.format(config['bucket_endpoint'], config['bucket_name'], project_bucket_key)

            runbiosimulations_id = biosimulators_utils.biosimulations.utils.run_simulation_project(
                name, project_url, 'cobrapy', project_id=project_id, purpose='academic', auth=auth)
            updated = str(update_times[model['model_bigg_id']])

        # output status
        status[model['model_bigg_id']] = {
            'created': status.get(model['model_bigg_id'], {}).get('created', str(update_times[model['model_bigg_id']])),
            'updated': updated,
            'objective': objective,
            'duration': duration,
            'runbiosimulationsId': runbiosimulations_id,
            'biosimulationsId': model['model_bigg_id'],
        }
        with open(config['status_filename'], 'w') as file:
            file.write(yaml.dump(status))

        thumbnails_curation_copy = copy.deepcopy(thumbnails_curation)
        for model in thumbnails_curation_copy.values():
            for thumbnail in model:
                thumbnail['filename'] = os.path.relpath(thumbnail['filename'], config['source_thumbnails_dirname'])
        with open(config['thumbnails_filename'], 'w') as file:
            file.write(yaml.dump(thumbnails_curation_copy))
